# Copyright 2008-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains functions to serialize L{TextBuffer} content

NOTE: This type of serialization should only be used for representation
of C{TextBuffer} contents in classes directly related to the C{TextBuffer}
and for testing purposes. All other parsing should us C{ParseTree} instead.
'''

import re
import logging

logger = logging.getLogger('zim.gui.pageview')


from zim.plugins import PluginManager
from zim.formats import IMAGE, OBJECT, ANCHOR, LINK, TAG, HEADING, LINE, TABLE
from zim.parse.tokenlist import TEXT, END, collect_until_end_token
from zim.parse.xml import simple_token_to_xml_dumper, simple_xml_to_token_parser

from .constants import PIXBUF_CHR, BULLETS_FROM_STOCK, ICON, BLOCK, LISTITEM
from .objectanchors import InsertedObjectAnchor, LineSeparatorAnchor
from .textbuffer import _is_inline_nesting_tag

__all__ = ('textbuffer_internal_insert_at_cursor', 'textbuffer_internal_serialize_range', 'TextBufferInternalContents')

_INTERNAL_ROOT = 'zim-textbuffer'
_INTERNAL_OBJECT_LIKE_TAGS = (IMAGE, ANCHOR, ICON, LINE) # These are not text ranges, but pixbuf / object anchor


from zim.parse.tokenlist import TokenBuilder
class MyBuilder(TokenBuilder):

	def data(self, data):
		self.text(data)


def _object_to_tag(objectanchor):
	# TODO - update object interface to do this without builder
	builder = MyBuilder()
	objectanchor.dump(builder)
	tokens = builder._tokens
	if tokens[0][0] == LINE:
		return [(LINE, None)]
	elif tokens[0][0] == TABLE:
		return tokens
	else:
		assert tokens[0][0] == OBJECT
		if len(tokens) > 3:
			data = ''.join(t[1] for t in tokens[1:-1])
			tokens = [tokens[0], (TEXT, data), tokens[-1]]
		return tokens


class TextBufferInternalContents():
	'''Container object for "raw" textbuffer content serialization

	The key difference between the content in this object and the content in
	a L{ParseTree} object is that this object is intended to exaxtly reproduce
	the L{TextBuffer} contents even if it contains semantic errors. This is 
	relevant specifically for undo/redo actions that should reproduce the 
	exact state of the buffer. The 	L{ParseTree} applies various filters and
	cleans up errors where possible and therefore is a much more stable and 
	clean interface to the document contents.

	NOTE: This type of serialization should only be used for representation
	of C{TextBuffer} contents in classes directly related to the C{TextBuffer}
	and for testing purposes. All other parsing should us C{ParseTree} instead.
	
	**The content of this class is not considered stable for outside APIs.**
	'''

	__slots__ = ('_data',)

	def __init__(self, data):
		self._data = data

	def __eq__(self, data):
		return self.__class__ == data.__class__ and self._data == data._data

	def __repr__(self):
		return '<%s %r>' % (self.__class__.__name__, self._data)

	def __str__(self):
		return self.to_xml()

	@classmethod
	def new_from_xml(cls, xml):
		'''Convert xml-like to object instance
		
		NOTE: Inteded for test cases only, do not use in the application - 
		not robust / un-save for external or mal-formed context
		'''
		data = []
		token_iter = simple_xml_to_token_parser(xml)
		for t in token_iter:
			if t[0] in (TEXT, END):
				data.append(t)
			else:
				data.append(t)
				if t[0] in _INTERNAL_OBJECT_LIKE_TAGS:
					# skip end tag
					end = next(token_iter)
					assert end == (END, t[0]), 'Expected %s got %s' % ((END, t[0]), end)

		if data[0][0] == _INTERNAL_ROOT:
			data = data[1:-1]
		return cls(data)

	def to_xml(self):
		'''Convert obejct contents to xml-like
		
		NOTE: Inteded for test cases only, do not use in the application - 
		not robust / un-save for external or mal-formed context
		'''	
		xml = simple_token_to_xml_dumper(
			self._data,
			tags_without_end_tag=_INTERNAL_OBJECT_LIKE_TAGS
		)
		return '<%s>%s</%s>' % (_INTERNAL_ROOT, xml, _INTERNAL_ROOT)



_is_zim_tag = lambda tag: hasattr(tag, 'zim_tag')


def textbuffer_internal_serialize_range(textbuffer, start, end):
	'''Function that returns a C{TextBufferInternalContents} object
	
	Basic serialization is in terms of TextTags, PixBufs and TextChildAnchors
	These are represented as tokens where TextTags have a start and an end token
	and text, pixbufs and anchors are represented as single tokens

	This method does not enforce zim's semantic rules becaues it is used for 
	direct representation of the buffer, including errors.

	The only logic it applies is the hierarchic nesting of TextTags to allow 
	a stable xml-like representation.
	'''
	textbuffer_data = []
	texttags_stack = [] # list of (texttag, zimtag)

	iter = start.copy()
	while iter.compare(end) == -1:
		open_tags = list(filter(_is_zim_tag, iter.get_tags()))
		if iter.equal(start):
			textbuffer_data.extend(
				_init_texttags_stack(texttags_stack, iter, open_tags) )
		else:
			textbuffer_data.extend(
				_update_texttags_stack(texttags_stack, iter, open_tags) )

		pixbuf = iter.get_pixbuf()
		anchor = iter.get_child_anchor()
		if pixbuf:
			assert hasattr(pixbuf, 'zim_type')
			textbuffer_data.append((pixbuf.zim_type, pixbuf.zim_attrib.copy()))
			iter.forward_char()
		elif anchor:
			assert isinstance(anchor, InsertedObjectAnchor)
			textbuffer_data.extend(_object_to_tag(anchor))
			iter.forward_char()
		else:
			# Find biggest slice without tags being toggled
			bound = iter.copy()
			toggled = []
			while not toggled:
				if not bound.is_end() and bound.forward_to_tag_toggle(None):
					# For some reason the not is_end check is needed to prevent an odd corner case infinite loop
					toggled = list(filter(_is_zim_tag, bound.get_toggled_tags(False) + bound.get_toggled_tags(True)))
						# breaks if toggle was indeed zim tag, else continue loop
				else:
					bound = end.copy() # just to be sure..
					break

			if bound.compare(end) == 1:
				bound = end.copy()

			# But limit slice to first pixbuf or embeddded widget
			text = iter.get_slice(bound)
			if PIXBUF_CHR in text[1:]:
				# Position 0 is a special case - we see this char, 
				# but get_pixbuf already returned None, so it must be taken literal
				i = 1 + text[1:].index(PIXBUF_CHR)
				bound = iter.copy()
				bound.forward_chars(i)
				text = text[:i]

			textbuffer_data.append((TEXT, text))
			iter = bound

	# close any open tags
	if texttags_stack:
		textbuffer_data.extend([(END, t[1]) for t in reversed(texttags_stack)])
	
	return TextBufferInternalContents(textbuffer_data)


def _init_texttags_stack(texttags_stack, iter, texttags):
	# Like _update_texttags_stack but for initalizing the stack
	# at the starting iter. This is subtly different from update because
	# sorting of inline tags also needs to look backward.
	# While updating this backward information is preserved on the stack
	tokens = []

	texttags.sort(key=lambda t: t.get_priority())
	if any(_is_inline_nesting_tag(t) for t in texttags):
		texttags = _sort_nesting_style_tags_for_init(iter, texttags)

	# Open tags
	for texttag in texttags:
		tag = texttag.zim_tag
		attrib = texttag.zim_attrib.copy() if texttag.zim_attrib else None
		tokens.append((tag, attrib))
		texttags_stack.append((texttag, tag))

	return tokens


def _update_texttags_stack(texttags_stack, iter, texttags):
	# This is a helper function that compares TextTags at a given iter with the 
	# stack of open tags. It updates the stack and returns open/close tokens
	# that reflect the change.
	#
	# It takes care of sorting stability of that tags which implies semantic
	# logic for how zim uses these tags.
	tokens = []

	texttags.sort(key=lambda t: t.get_priority())
	if any(_is_inline_nesting_tag(t) for t in texttags):
		texttags = _sort_nesting_style_tags_for_update(iter, texttags, [t[0] for t in texttags_stack])

	# Compare tags with stack
	i = 0
	while i < len(texttags) and i < len(texttags_stack) \
		and texttags[i] == texttags_stack[i][0]:
			i += 1

	# Close tags
	while len(texttags_stack) > i:
		tokens.append((END, texttags_stack[-1][1]))
		texttags_stack.pop()

	# Open tags
	for texttag in texttags[i:]:
		tag = texttag.zim_tag
		attrib = texttag.zim_attrib.copy() if texttag.zim_attrib else None
		tokens.append((tag, attrib))
		texttags_stack.append((texttag, tag))

	return tokens


def _sort_nesting_style_tags_for_init(iter, new_texttags):
	new_block, new_nesting, new_leaf = _split_nesting_style_tags(new_texttags)

	def tag_start_close_pos(tag):
		if iter.starts_tag(tag):
			start = iter.get_offset()
		else:
			my_iter = iter.copy()
			my_iter.backward_to_tag_toggle(tag)
			start = my_iter.get_offset()

		my_iter = iter.copy()
		my_iter.forward_to_tag_toggle(tag)
		end = my_iter.get_offset()
		return (start, end)

	new_nesting.sort(key=tag_start_close_pos, reverse=True)

	return new_block + new_nesting + new_leaf


def _sort_nesting_style_tags_for_update(iter, new_texttags, open_texttags):
	# This helper function figures out the sorting of style tags that have the same priority.
	# Tags that start earlier obviously go first. Of all tags opening at the same time, those 
	# which run longest get opened first - this way we optimize for longest stretch without breaks.
	new_block, new_nesting, new_leaf = _split_nesting_style_tags(new_texttags)
	open_block, open_nesting, open_leaf = _split_nesting_style_tags(open_texttags)
	sorted_new_nesting = []

	# First prioritize open tags - these are sorted already
	if new_block == open_block:
		for tag in open_nesting:
			if tag in new_nesting:
				i = new_nesting.index(tag)
				sorted_new_nesting.append(new_nesting.pop(i))
			else:
				break

	# Then sort by length untill closing all tags that open at the same time
	def tag_close_pos(tag):
		my_iter = iter.copy()
		my_iter.forward_to_tag_toggle(tag)
		return my_iter.get_offset()

	new_nesting.sort(key=tag_close_pos, reverse=True)
	sorted_new_nesting += new_nesting

	return new_block + sorted_new_nesting + new_leaf


def _split_nesting_style_tags(texttags):
	# Helper function to split a set of texttags in block level, nesting and leaf tags
	# This applies semantic knowledge on the tag interpretation
	block, nesting = [], []
	while texttags and not _is_inline_nesting_tag(texttags[0]):
		block.append(texttags.pop(0))
	while texttags and _is_inline_nesting_tag(texttags[0]):
		nesting.append(texttags.pop(0))
	return block, nesting, texttags


def textbuffer_internal_insert_at_cursor(textbuffer, data):
	'''Function that inserts L{TextBufferInternalContents} into a L{TextBuffer}'''
	interactive = False
	textbuffer.emit('begin-insert-tree', interactive)

	textbuffer._editmode_tags = [] # low level insert method, no carry over of open tags
	texttag_stack = []
	
	token_iter = iter(data._data)
	for t in token_iter:
		if t[0] == TEXT:
			textbuffer.insert_at_cursor(t[1])
		
		# Pixbuf types
		elif t[0] == IMAGE:
			textbuffer.insert_image_at_cursor(None, **t[1]) # TODO - should we use _file to store actual file object in attrib
		elif t[0] == ANCHOR:
			textbuffer.insert_anchor_at_cursor(**t[1])
		elif t[0] == ICON:
			bullet = BULLETS_FROM_STOCK[t[1]['stock']] # TODO low level function that takes stock
			textbuffer._insert_bullet_at_cursor(bullet)

		# InsertedObject
		elif t[0] == LINE:
			anchor = LineSeparatorAnchor()
			textbuffer.insert_objectanchor_at_cursor(anchor)
		elif t[0] == OBJECT:
			attrib = t[1].copy()
			data = ''
			for t in token_iter:
				if t[0] == TEXT:
					data += t[1]
				else:
					assert t == (END, OBJECT), 'Expected end of object, got: %r' % t
					break
			objecttype, model = textbuffer._get_objecttype_and_model_for_object(attrib, data or None)
			textbuffer._insert_object_model_at_cursor(objecttype, model)
		elif t[0] == TABLE:
			try:
				objecttype = PluginManager.insertedobjects['table']
			except KeyError:
				# HACK - if table plugin is not loaded - show table as plain text
				from zim.formats import ParseTree
				tree = ParseTree.new_from_tokens(tokens)
				lines = tree.dump('wiki')
				objecttype, model = textbuffer._get_objecttype_and_model_for_object({'type': 'table'}, ''.join(lines))
				textbuffer._insert_object_model_at_cursor(objecttype, model)
			else:
				tokens = collect_until_end_token(token_iter, TABLE)
				tokens.insert(0, t)
				model = objecttype.model_from_tokens(tokens)
				textbuffer._insert_object_model_at_cursor(objecttype, model)

		# TextTag types
		elif t[0] == END:
			assert texttag_stack and t[1] == texttag_stack[-1][0], 'Inconsistend closing tag: %r' % (t, )
			x, texttag = texttag_stack.pop()
			textbuffer._editmode_tags = [tt for tt in textbuffer._editmode_tags if tt is not texttag]
			if texttag.zim_tag == LINK:
				_postprocess_insert_link_tag(textbuffer.get_insert_iter(), texttag)
		else:
			if t[0] == LINK:
				texttag = _get_link_texttag(textbuffer.get_insert_iter())
				if texttag and texttag.zim_attrib == t[1]:
					pass # Continue open link tag
				else:
					texttag = textbuffer._create_link_tag('', **t[1])
			elif t[0] == TAG:
				texttag = textbuffer._create_tag_tag(None)
			elif t[0] == HEADING:
				texttag = textbuffer.get_tag_table().lookup('style-h' + str(t[1]['level']))
			elif t[0] == BLOCK:
				texttag = textbuffer._get_indent_tag(t[1]['indent'])
				# TODO: note from original insert function: We don't set the LTR / RTL direction here
				# instead we update all indent tags after the full
				# insert is done.
			elif t[0] == LISTITEM:
				texttag = textbuffer._get_indent_tag(t[1]['indent'], t[1]['style'])
				# TODO: note from original insert function: We don't set the LTR / RTL direction here
				# instead we update all indent tags after the full
				# insert is done.
			else:
				texttag = textbuffer.get_tag_table().lookup('style-' + t[0])

			if not texttag:
				raise AssertionError('Unknown internal data element: %r' % (t,))			
			texttag_stack.append((t[0], texttag))
			textbuffer._editmode_tags.append(texttag)

	if texttag_stack:
		logger.warn('Tags not closed on insert: %r' % texttag_stack)

	textbuffer.update_editmode() # Ensure we leave buffer ready for interactive editing
	textbuffer.emit('end-insert-tree')


def _get_link_texttag(startiter):
	# Like Textbuffer.get_link_tag() but with right gravity instead of left gravity
	texttags = list(startiter.get_tags()) + list(startiter.get_toggled_tags(False))
	for texttag in sorted(texttags, key=lambda i: i.get_priority()):
			if hasattr(texttag, 'zim_tag') and texttag.zim_tag == 'link':
				return texttag
	else:
		return None	


def _postprocess_insert_link_tag(enditer, texttag):
	# If the "href" attribute matches the text of the link it should
	# be set to `None` to flag that the href is "dynamic" and should 
	# change with the text. Otherwise the href is fixed regardless of 
	# text editing.
	
	# TODO: this logic is copy of logic in _create_link_tag -- refactor for re-use
	href = texttag.zim_attrib['href']
	if not href or href.isspace():
		texttag.zim_attrib['href'] = None
	else:
		startiter = enditer.copy()
		startiter.backward_to_tag_toggle(texttag)
		text = startiter.get_slice(enditer)
		if href == text:
			texttag.zim_attrib['href'] = None



# Copyright 2008-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains the classes which implement the "find" functionality

"Find" means a match of text in the text buffer of a given page or inline
object. This differs from "search" which means a more complex query over
many pages.
'''


from gi.repository import GObject
from gi.repository import Gtk

import re
import logging
import itertools

from typing import Optional

from zim.gui.widgets import Dialog, IconButton, InputEntry

from .constants import *


logger = logging.getLogger('zim.gui.pageview')


# Tag styles
FIND_HIGHLIGHT_TAG = 'find-highlight'
FIND_MATCH_TAG = 'find-match'


# Query options
FIND_CASE_SENSITIVE = 1 #: Constant to find case sensitive
FIND_WHOLE_WORD = 2 #: Constant to find whole words only
FIND_REGEX = 4 #: Constant to find based on regexes


class FindQuery():
	'''Object which represents a query for the find functionality

	"Find" is a direct match of text content in the buffer, therefore
	all queries can be represented as regular expressions.
	'''

	def __init__(self, string: str, flags: int = 0):
		self.string = string
		self.flags = flags
		self.regex = self._compile_regex(string, flags)

	def __eq__(self, other):
		return isinstance(other, self.__class__) and (self.string, self.flags) == (other.string, other.flags)

	def __bool__(self):
		return bool(self.string)

	@staticmethod
	def _compile_regex(string, flags):
		assert isinstance(string, str)
		assert isinstance(flags, int)

		if not flags & FIND_REGEX:
			string = re.escape(string)

		if flags & FIND_WHOLE_WORD:
			string = '\\b' + string + '\\b'

		if flags & FIND_CASE_SENSITIVE:
			return re.compile(string, re.U)
		else:
			return re.compile(string, re.U | re.I)


class FindInterface():
	'''Interface definition for the TextBuffer and inserted objects

	The interface implies a statefull model where a match is highlighted and this highlighting
	is reset on the next call or on L{find_clear()}. In addition the L{find_highlight_all()} method
	will highlight all matches. All of the operations L{find_next()}, L{find_previous()},
	L{find_replace_next()} and L{find_replace_all()} will reset the highlighting unless the query is
	the same as provided to L{find_highlight_all()}. This implies the object remembers the query
	used for highlighting.
	Vice versa, a change to the highlighting will reset any matches in the buffer or object.
	'''

	def find_has_active_match(self) -> bool:
		'''True if there is an active match on the current cursor location'''
		raise NotImplementedError

	def find_next(self, query: FindQuery, wrap: bool = True) -> bool:
		'''Go to the next match and highlight it
		If a match happens directly at the current cursor position, it will only match if
		not yet highlighted as a match.
		@param query: string with search query
		@param wrap: if C{True}, wrap around at the end of the buffer and match from the start
		@returns: C{True} if successful
		'''
		raise NotImplementedError

	def find_next_from_start(self, query: FindQuery) -> bool:
		'''Like C{find_next()} but intended for nested objects
		Behavior is statefull:
		If there is a match in the object, continue to find the next match from there;
		if there is no match in the object, look for a match from the start.
		Should not wrap around.
		'''
		raise NotImplementedError

	def find_previous(self, query: FindQuery, wrap: bool = True) -> bool:
		'''Go to the previous match and highlight it
		If a match happens directly at the current cursor position, it will only match if
		not yet highlighted as a match.
		@param query: string with search query
		@param wrap: if C{True}, wrap around at the start of the buffer and match from the end
		@returns: C{True} if successful
		'''
		raise NotImplementedError

	def find_previous_from_end(self, query: FindQuery) -> bool:
		'''Like C{find_previous()} but intended for nested objects
		Behavior is statefull:
		If there is a match in the object, continue to find the previous match from there;
		if there is no match in the object, look for a match from the start.
		Should not wrap around.
		'''
		raise NotImplementedError

	def find_highlight_all(self, query: FindQuery):
		'''Highlight all matches of the query
		@param query: string with search query
		'''
		raise NotImplementedError

	def find_clear():
		'''Clear all find highlighting and matches'''
		raise NotImplementedError

	def find_replace_at_cursor(self, query: FindQuery, replacement: str) -> bool:
		'''Replace match at cursor

		Replace a match but only if it occurs directly at the cursor. Typically this is combined
		with a find operation such that the user sees the match before pressing "replace" button.
		In case of a regex find and replace the string will be expanded with terms from the regex.

		@param query: string with search query
		@param replacement: new text to replace the match
		@returns: C{True} if successful
		'''
		raise NotImplementedError

	def find_replace_all(self, query: FindQuery, replacement: str) -> bool:
		'''Replace all matches

		In case of a regex find and replace the string will be expanded with terms from the regex.

		@param query: string with search query
		@param replacement: new text to replace the match
		@returns: C{True} if successful
		'''
		raise NotImplementedError


def _my_chain_reversed(*iters):
	# Bit inefficient because we need to iterate all to get the last match
	# but at least do it by iterator instead of all at once
	# Assume `find_previous()` used less often than `find_next()` and `find_highlight_all()`
	for myiter in iters:
		results = list(myiter)
		results.reverse()
		for r in results:
			yield r


class TextBufferFindMixin(FindInterface):
	'''Implemenation of the L{FindInterface} which can be combined with either a
	standard C{Gtk.TextBuffer}s or a zim specific L{TextBuffer}
	'''

	# Methods specific for the TextBuffer implemenation

	def __init__(self):
		if hasattr(self, 'tag_styles'):
			self._find_match_tag = self.create_tag(None, **self.tag_styles[FIND_MATCH_TAG])
			self._find_highlight_tag = self.create_tag(None, **self.tag_styles[FIND_HIGHLIGHT_TAG])
		else:
			# Bit of a hack to re-use same style definition
			from .textbuffer import TextBuffer
			self._find_match_tag = self.create_tag(None, **TextBuffer.tag_styles[FIND_MATCH_TAG])
			self._find_highlight_tag = self.create_tag(None, **TextBuffer.tag_styles[FIND_HIGHLIGHT_TAG])

		self._find_signals = ()
		self._find_highlight_all_query = None

	def _find_all_in_range(self, query, start, end):
		# Generator for all matches and objects in range
		# objects are yielded if they support FindInterface, but not checked for matches

		if query.flags & FIND_WHOLE_WORD:
			if start.inside_word() and not start.starts_word():
				start.forward_word_end()

			if end.inside_word() and not end.starts_word():
				end.backward_word_start()

		# See if there are objects supporting FindInterface
		text = start.get_slice(end)
		iobjects = []
		if hasattr(self, 'get_objectanchor'): # specific for zim TextBuffer
			i = text.find(PIXBUF_CHR)
			while i >= 0:
				iter = start.copy()
				iter.forward_chars(i)
				anchor = self.get_objectanchor(iter)
				if anchor:
					iobjects.append((i, anchor))
				i = text.find(PIXBUF_CHR, i+1)

		# Check for text matches in between the objects and yield results
		pos = 0
		iobjects.append((len(text), None)) # ensure last part also searched
		for endpos, iobject in iobjects:
			for match in query.regex.finditer(text, pos, endpos):
				mstart, mend = start.copy(), start.copy()
				mstart.forward_chars(match.start())
				mend.forward_chars(match.end())
				yield mstart, mend, match

			if isinstance(iobject, FindInterface):
				mstart, mend = start.copy(), start.copy()
				mstart.forward_chars(endpos)
				mend.forward_chars(endpos+1)
				yield mstart, mend, iobject

			pos = endpos + 1

	def _find_set_match(self, start, end):
		for id in self._find_signals:
			self.disconnect(id)

		self.apply_tag(self._find_match_tag, start, end)
		self.select_range(start, end)

		self._find_signals = tuple(
			self.connect(s, self._find_on_buffer_changed)
				for s in ('mark-set', 'changed'))

	def _find_on_buffer_changed(self, *a):
		if len(a) > 2 and isinstance(a[2], Gtk.TextMark) \
		and a[2] is not self.get_insert():
			# mark-set signal, but not for cursor
			return

		for id in self._find_signals:
			self.disconnect(id)
		self._find_signals = ()
		self.remove_tag(self._find_match_tag, *self.get_bounds())

	def _find_clear_on_new_query(self, query):
		if self._find_highlight_all_query and not self._find_highlight_all_query == query:
			self.find_clear()
		else:
			# just clear previous match
			self.remove_tag(self._find_match_tag, *self.get_bounds())

	# Implementation of FindInterface based on above methods

	def find_has_active_match(self):
		iter = self.get_iter_at_mark(self.get_insert())
		return self._find_match_tag in iter.get_toggled_tags(toggled_on=True)

	def find_next(self, query, wrap=True):
		assert isinstance(query, FindQuery)

		iter = self.get_iter_at_mark(self.get_insert())
		if self.find_has_active_match():
			iter.forward_char() # avoid match at current location
		elif hasattr(self, 'get_objectanchor'): # specific for zim TextBuffer
			# maybe last action was a find_previous with an object match,
			# check whether we are behind an object with an active match
			tmp = iter.copy()
			tmp.backward_char()
			anchor = self.get_objectanchor(tmp)
			if anchor and isinstance(anchor, FindInterface) \
				and anchor.find_has_active_match() and anchor.find_next(query, wrap=False):
					self._find_clear_on_new_query(query)
					return True

		self._find_clear_on_new_query(query)

		bstart, bend = self.get_bounds()
		if wrap:
			matches = itertools.chain(
				self._find_all_in_range(query, iter, bend),
				self._find_all_in_range(query, bstart, iter)
			)
		else:
			matches = self._find_all_in_range(query, iter, bend)

		for mstart, mend, match in matches:
			if isinstance(match, FindInterface):
				# Object match
				if match.find_next_from_start(query):
					self.place_cursor(mstart)
					return True
			else:
				# Regex text match
				self._find_set_match(mstart, mend)
				return True
		else:
			return False

	def find_next_from_start(self, query):
		if not self.find_has_active_match():
			self.place_cursor(self.get_start_iter())
		return self.find_next(query, wrap=False)

	def find_previous(self, query, wrap=True):
		assert isinstance(query, FindQuery)

		if hasattr(self, 'get_objectanchor'): # specific for zim TextBuffer
			# maybe last action was a find_next with an object match,
			# check whether we are in front of an object with an active match
			anchor = self.get_objectanchor_at_cursor()
			if anchor and isinstance(anchor, FindInterface) \
				and anchor.find_has_active_match() and anchor.find_previous(query, wrap=False):
					self._find_clear_on_new_query(query)
					return True

		self._find_clear_on_new_query(query)

		iter = self.get_iter_at_mark(self.get_insert())
		bstart, bend = self.get_bounds()
		if wrap:
			matches = _my_chain_reversed(
				self._find_all_in_range(query, bstart, iter),
				self._find_all_in_range(query, iter, bend)
			)
		else:
			matches = list(self._find_all_in_range(query, bstart, iter))
			matches.reverse()

		for mstart, mend, match in matches:
			if isinstance(match, FindInterface):
				# Object match
				if match.find_previous_from_end(query):
					self.place_cursor(mend)
					return True
			else:
				# Regex text match
				self._find_set_match(mstart, mend)
				return True
		else:
			return False

	def find_previous_from_end(self, query):
		if not self.find_has_active_match():
			self.place_cursor(self.get_end_iter())
		return self.find_previous(query, wrap=False)

	def find_highlight_all(self, query):
		assert isinstance(query, FindQuery)
		self.find_clear()
		self._find_highlight_all_query = query

		bstart, bend = self.get_bounds()
		for mstart, mend, match in self._find_all_in_range(query, bstart, bend):
			if isinstance(match, FindInterface):
				# Object match
				match.find_highlight_all(query)
			else:
				# Regex text match
				self.apply_tag(self._find_highlight_tag, mstart, mend)

	def find_clear(self):
		self.remove_tag(self._find_match_tag, *self.get_bounds())
		self.remove_tag(self._find_highlight_tag, *self.get_bounds())
		self._find_highlight_all_query = None
		if hasattr(self, 'list_objectanchors'): # specific for zim TextBuffer
			for _, anchor in self.list_objectanchors():
				if isinstance(anchor, FindInterface):
					anchor.find_clear()

	def find_replace_at_cursor(self, query, replacement):
		assert isinstance(query, FindQuery)
		self._find_clear_on_new_query(query)

		iter = self.get_iter_at_mark(self.get_insert())
		end = iter.copy()
		end.forward_to_line_end()
		mstart, mend, match = next(self._find_all_in_range(query, iter, end), (None, None, None))
		if match and mstart.equal(iter):
			if isinstance(match, FindInterface):
				# Object match
				return match.find_replace_at_cursor(query, replacement)
			else:
				# Regex text match
				if query.flags & FIND_REGEX:
					try:
						replacement = match.expand(replacement)
					except:
						logger.exception('error in regex expansion')

				offset = mstart.get_offset()

				self.begin_user_action()
				self.select_range(mstart, mend) # ensure editmode logic is used
				self.delete(mstart, mend)
				self.insert_at_cursor(replacement)
				self.end_user_action()

				rstart = self.get_iter_at_offset(offset)
				rend = self.get_iter_at_offset(offset + len(replacement))
				self.select_range(rstart, rend)

				return True
		else:
			return False # No match at cursor

	def find_replace_all(self, query, replacement):
		# NOTE: Avoid looping when replace value matches query
		assert isinstance(query, FindQuery)
		self._find_clear_on_new_query(query)

		matches = []
		for mstart, mend, match in self._find_all_in_range(query, *self.get_bounds()):
			if isinstance(match, FindInterface):
				# Object match
				match.find_replace_all(query, replacement)
			else:
				# Regex text match
				my_replacement = replacement
				if query.flags & FIND_REGEX:
					try:
						my_replacement = match.expand(replacement)
					except:
						logger.exception('error in regex expansion')

				matches.append((mstart.get_offset(), mend.get_offset(), my_replacement))

		matches.reverse() # work our way backward to keep offsets valid

		self.begin_user_action()
		for startoff, endoff, string in matches:
			start = self.get_iter_at_offset(startoff)
			end = self.get_iter_at_offset(endoff)
			self.delete(start, end)
			start = self.get_iter_at_offset(startoff)
			self.insert(start, string)
		self.end_user_action()

		return len(matches) > 0


FIND_HAS_MATCH = 1
FIND_HAS_HIGHLIGHT = 2
FIND_HAS_MATCH_CSS_CLASS = 'find_match'
FIND_HAS_HIGHLIGHT_CSS_CLASS = 'find_highlight'


class PluginInsertedObjectFindMixin(FindInterface):
	'''Implemenation of the L{FindInterface} for inserted objects from a plugin

	This class checks whether the model of the object supports the L{FindInterface} - e.g. if it is
	a nested C{TextBuffer} and if so will pass through the method calls.
	Otherwise this class treats the objects as "atomic" matches based on the object data and does not
	support replace.
	'''

	# Methods implementing logic specific for simple object

	_find_match_highlight_state = 0 # use class attrib to init instance attrib
	_find_match_query = None # use class attrib to init instance attrib

	def _find_set_match_highlight_state(self, state: int, query: Optional[FindQuery]):
		self._find_match_highlight_state = state
		self._find_match_query = query
		for widget in self.widgets:
			widget = widget._vbox # XXX
			context = widget.get_style_context()
			context.remove_class(FIND_HAS_MATCH_CSS_CLASS)
			context.remove_class(FIND_HAS_HIGHLIGHT_CSS_CLASS)
			if state & FIND_HAS_MATCH:
				context.add_class(FIND_HAS_MATCH_CSS_CLASS)
			elif state & FIND_HAS_HIGHLIGHT:
				context.add_class(FIND_HAS_HIGHLIGHT_CSS_CLASS)
			# else nothing

	def _find_simple_match(self, query: FindQuery) -> bool:
		'''Returns C{True} if current object matches the query'''
		if hasattr(self.objectmodel, 'find_simple_match'):
			return self.objectmodel.find_simple_match(query)
		else:
			attrib, data = self.objecttype.data_from_model(self.objectmodel)
			return (query.regex.search(data) is not None) if data else False

	# Implementation of FindInterface based on above methods

	def find_has_active_match(self):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_has_active_match()
		else:
			return self._find_match_highlight_state & FIND_HAS_MATCH

	def find_next(self, query, wrap=True):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_next(query, wrap)
		else:
			return self._find_any(query)

	def find_next_from_start(self, query):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_next_from_start(query)
		else:
			return self._find_any(query)

	def find_previous(self, query, wrap=True):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_previous(query, wrap)
		else:
			return self._find_any(query)

	def find_previous_from_end(self, query):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_previous_from_end(query)
		else:
			return self._find_any(query)

	def _find_any(self, query):
		if self._find_match_highlight_state & FIND_HAS_MATCH \
			and self._find_match_query == query:
				ok = False # Skip if already matched
				self._find_set_match_highlight_state(self._find_match_highlight_state ^ FIND_HAS_MATCH, query)
		else:
			ok = self._find_simple_match(query)
			if ok:
				if self._find_match_highlight_state & FIND_HAS_HIGHLIGHT \
					and self._find_match_query == query:
						new_state = FIND_HAS_MATCH | FIND_HAS_HIGHLIGHT
				else:
					new_state = FIND_HAS_MATCH
				self._find_set_match_highlight_state(new_state, query)
			else:
				self._find_set_match_highlight_state(0, None)

		return ok

	def find_highlight_all(self, query):
		if isinstance(self.objectmodel, FindInterface):
			ok = self.objectmodel.find_highlight_all(query)
		else:
			if self._find_simple_match(query):
				self._find_set_match_highlight_state(FIND_HAS_HIGHLIGHT, query)
			else:
				self._find_set_match_highlight_state(0, None)

	def find_clear(self):
		if isinstance(self.objectmodel, FindInterface):
			self.objectmodel.find_clear()
		else:
			self._find_set_match_highlight_state(0, None)

	def find_replace_at_cursor(self, query, replacement):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_replace_at_cursor(query, replacement)
		else:
			return False # Not supported

	def find_replace_all(self, query, replacement):
		if isinstance(self.objectmodel, FindInterface):
			return self.objectmodel.find_replace_all(query, replacement)
		else:
			return False # Not supported


class FindWidget(object):
	'''Base class for L{FindBar} and L{FindAndReplaceDialog}'''

	def __init__(self, textview):
		self.textview = textview

		self.find_entry = InputEntry(allow_whitespace=True)
		self.find_entry.connect_object(
			'changed', self.__class__.on_find_entry_changed, self)
		self.find_entry.connect_object(
			'activate', self.__class__.on_find_entry_activate, self)

		self.next_button = Gtk.Button.new_with_mnemonic(_('_Next'))
			# T: button in find bar and find & replace dialog
		self.next_button.connect_object(
			'clicked', self.__class__.find_next, self)
		self.next_button.set_sensitive(False)

		self.previous_button = Gtk.Button.new_with_mnemonic(_('_Previous'))
			# T: button in find bar and find & replace dialog
		self.previous_button.connect_object(
			'clicked', self.__class__.find_previous, self)
		self.previous_button.set_sensitive(False)

		self.case_option_checkbox = Gtk.CheckButton.new_with_mnemonic(_('Match _case'))
			# T: checkbox option in find bar and find & replace dialog
		self.case_option_checkbox.connect_object(
			'toggled', self.__class__.on_find_entry_changed, self)

		self.word_option_checkbox = Gtk.CheckButton.new_with_mnemonic(_('Whole _word'))
			# T: checkbox option in find bar and find & replace dialog
		self.word_option_checkbox.connect_object(
			'toggled', self.__class__.on_find_entry_changed, self)

		self.regex_option_checkbox = Gtk.CheckButton.new_with_mnemonic(_('_Regular expression'))
			# T: checkbox option in find bar and find & replace dialog
		self.regex_option_checkbox.connect_object(
			'toggled', self.__class__.on_find_entry_changed, self)

		self.highlight_checkbox = Gtk.CheckButton.new_with_mnemonic(_('_Highlight'))
			# T: checkbox option in find bar and find & replace dialog
		self.highlight_checkbox.connect_object(
			'toggled', self.__class__.on_highlight_toggled, self)

	def _get_query(self):
		string = self.find_entry.get_text()
		flags = 0
		if self.case_option_checkbox.get_active():
			flags = flags | FIND_CASE_SENSITIVE
		if self.word_option_checkbox.get_active():
			flags = flags | FIND_WHOLE_WORD
		if self.regex_option_checkbox.get_active():
			flags = flags | FIND_REGEX
		return FindQuery(string, flags)

	def on_find_entry_changed(self, match_at_cursor=True):
		query = self._get_query()
		buffer = self.textview.get_buffer()
		if match_at_cursor:
			buffer.find_clear() # prevent skipping past current cursor if cursor matches

		if not query:
			ok = False
			self.find_entry.set_input_valid(True)
			buffer.find_clear() # also clear highlight (if not done yet)
		else:
			ok = buffer.find_next(query)
			self.find_entry.set_input_valid(ok)
			if self.highlight_checkbox.get_active():
				buffer.find_highlight_all(query)

		for button in (self.next_button, self.previous_button):
			button.set_sensitive(ok)

		if ok:
			self.textview.scroll_to_mark(buffer.get_insert(), SCROLL_TO_MARK_MARGIN, False, 0, 0)

	def on_find_entry_activate(self):
		self.on_find_entry_changed(match_at_cursor=False)

	def on_highlight_toggled(self):
		buffer = self.textview.get_buffer()
		if self.highlight_checkbox.get_active():
			buffer.find_highlight_all(self._get_query())
		else:
			buffer.find_clear()

	def find(self, string, flags=0, highlight=False):
		self.find_entry.set_text(string)
		self.case_option_checkbox.set_active(flags & FIND_CASE_SENSITIVE)
		self.word_option_checkbox.set_active(flags & FIND_WHOLE_WORD)
		self.regex_option_checkbox.set_active(flags & FIND_REGEX)
		self.highlight_checkbox.set_active(highlight)
		self.on_find_entry_changed()

	def find_next(self):
		buffer = self.textview.get_buffer()
		ok = buffer.find_next(self._get_query())
		if ok:
			self.textview.scroll_to_mark(buffer.get_insert(), SCROLL_TO_MARK_MARGIN, False, 0, 0)

	def find_previous(self):
		buffer = self.textview.get_buffer()
		ok = buffer.find_previous(self._get_query())
		if ok:
			self.textview.scroll_to_mark(buffer.get_insert(), SCROLL_TO_MARK_MARGIN, False, 0, 0)


class FindBar(FindWidget, Gtk.ActionBar):
	'''Bar to be shown below the TextView for find functions'''

	# Ideas:
	# - support smaller buttons
	# - allow box to shrink further by putting buttons in a menu when to little space

	def __init__(self, textview):
		GObject.GObject.__init__(self)
		FindWidget.__init__(self, textview)

		self.pack_start(Gtk.Label(_('Find') + ': '))
			# T: label for input in find bar on bottom of page
		self.pack_start(self.find_entry)
		self.pack_start(self.previous_button)
		self.pack_start(self.next_button)
		self.pack_start(self.case_option_checkbox)
		self.pack_start(self.highlight_checkbox)

		close_button = IconButton(Gtk.STOCK_CLOSE, relief=False, size=Gtk.IconSize.MENU)
		close_button.connect_object('clicked', self.__class__.hide, self)
		self.pack_end(close_button)

	def grab_focus(self):
		self.find_entry.grab_focus()

	def show(self):
		if not self.get_visible():
			self.set_no_show_all(False)
			self.show_all()
			self.on_find_entry_changed(match_at_cursor=True)
		# else avoid changing state

	def hide(self):
		Gtk.ActionBar.hide(self)
		self.set_no_show_all(True)
		buffer = self.textview.get_buffer()
		buffer.find_clear()
		self.textview.grab_focus()


class FindAndReplaceDialog(FindWidget, Dialog):
	'''Dialog for find and replace'''

	def __init__(self, parent, textview):
		Dialog.__init__(self, parent,
			_('Find and Replace'), buttons=Gtk.ButtonsType.CLOSE) # T: Dialog title
		FindWidget.__init__(self, textview)

		hbox = Gtk.HBox(spacing=12)
		hbox.set_border_width(12)
		self.vbox.add(hbox)

		vbox = Gtk.VBox(spacing=5)
		hbox.pack_start(vbox, True, True, 0)

		label = Gtk.Label(label=_('Find what') + ': ')
			# T: input label in find & replace dialog
		label.set_alignment(0.0, 0.5)
		vbox.add(label)
		vbox.add(self.find_entry)
		vbox.add(self.case_option_checkbox)
		vbox.add(self.word_option_checkbox)
		vbox.add(self.regex_option_checkbox)
		vbox.add(self.highlight_checkbox)

		label = Gtk.Label(label=_('Replace with') + ': ')
			# T: input label in find & replace dialog
		label.set_alignment(0.0, 0.5)
		vbox.add(label)
		self.replace_entry = InputEntry(allow_whitespace=True)
		vbox.add(self.replace_entry)

		self.bbox = Gtk.ButtonBox(orientation=Gtk.Orientation.VERTICAL)
		self.bbox.set_layout(Gtk.ButtonBoxStyle.START)
		self.bbox.set_spacing(5)
		hbox.pack_start(self.bbox, False, False, 0)
		self.bbox.add(self.next_button)
		self.bbox.add(self.previous_button)

		replace_button = Gtk.Button.new_with_mnemonic(_('_Replace'))
			# T: Button in search & replace dialog
		replace_button.connect_object('clicked', self.__class__.replace, self)
		self.bbox.add(replace_button)

		all_button = Gtk.Button.new_with_mnemonic(_('Replace _All'))
			# T: Button in search & replace dialog
		all_button.connect_object('clicked', self.__class__.replace_all, self)
		self.bbox.add(all_button)

	def set_input(self, **inputs):
		# Hide implementation for test cases
		for key, value in list(inputs.items()):
			if key == 'query':
				self.find_entry.set_text(value)
			elif key == 'replacement':
				self.replace_entry.set_text(value)
			else:
				raise ValueError

	def replace(self):
		string = self.replace_entry.get_text()
		buffer = self.textview.get_buffer()
		query = self._get_query()
		if buffer.find_replace_at_cursor(query, string):
			buffer.find_next(query)

	def replace_all(self):
		string = self.replace_entry.get_text()
		buffer = self.textview.get_buffer()
		buffer.find_replace_all(self._get_query(), string)

	def do_response(self, id):
		Dialog.do_response(self, id)
		buffer = self.textview.get_buffer()
		buffer.find_clear()





# Copyright 2009-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

from gi.repository import GObject
from gi.repository import Gio
from gi.repository import Gtk

import logging

from zim.plugins import PluginClass
from zim.config import data_file, ConfigManager
from zim.main import GtkCommand
from zim.signals import SignalEmitter
from zim.notebook import get_notebook_list, NotebookInfo, NotebookInfoList

from zim.gui.mainwindow import MainWindowExtension

# Try if we are on Ubunutu with app-indicator support
try:
	import gi
	gi.require_version('AppIndicator3', '0.1')
	from gi.repository import AppIndicator3 as AppIndicator
except:
	AppIndicator = None


logger = logging.getLogger('zim.plugins.trayicon')



GLOBAL_TRAYICON = None
# The trayicon is global because you should only have one per process
# so in a sense it is an external resource represented by a global
# state.


def set_global_trayicon(classic=False):
	global GLOBAL_TRAYICON

	if AppIndicator and not classic:
		cls = AppIndicatorTrayIcon
	else:
		cls = StatusIconTrayIcon

	if GLOBAL_TRAYICON and isinstance(GLOBAL_TRAYICON, cls):
		return None

	if GLOBAL_TRAYICON:
		GLOBAL_TRAYICON.destroy()

	GLOBAL_TRAYICON = cls()
	return GLOBAL_TRAYICON


class TrayIconPluginCommand(GtkCommand):
	'''Class to handle "zim --plugin trayicon" allows starting zim in
	the background and only show a trayicon
	'''

	def run(self):
		preferences = ConfigManager.preferences['TrayIconPlugin']
		preferences.setdefault('classic', False)
		return set_global_trayicon(preferences['classic'])


class TrayIconPlugin(PluginClass):

	plugin_info = {
		'name': _('Tray Icon'), # T: plugin name
		'description': _('''\
This plugin adds a tray icon for quick access.

This plugin depends on Gtk+ version 2.10 or newer.

This is a core plugin shipping with zim.
'''), # T: plugin description
		'author': 'Jaap Karssenberg',
		'help': 'Plugins:Tray Icon',
	}

	plugin_preferences = (
		# key, type, label, default
		('classic', 'bool', _('Classic trayicon,\ndo not use new style status icon on Ubuntu'), False), # T: preferences option
	)

	@classmethod
	def check_dependencies(klass):
		return (True, [
			('Unity appindicator', bool(AppIndicator), False),
		])

	def __init__(self):
		PluginClass.__init__(self)
		self.preferences.connect('changed', self.on_preferences_changed)

	def on_preferences_changed(self, preferences):
		if GLOBAL_TRAYICON:
			# Only refresh once someone else loaded it
			application = GLOBAL_TRAYICON.get_application() if hasattr(GLOBAL_TRAYICON, 'get_application') else None
			trayicon = set_global_trayicon(self.preferences['classic'])
			if trayicon and application:
				application.add_window(trayicon)


class TrayIconMainWindowExtension(MainWindowExtension):

	def __init__(self, plugin, window):
		MainWindowExtension.__init__(self, plugin, window)
		self.window.hideonclose = True
		icon = set_global_trayicon(plugin.preferences['classic'])
		application = window.get_application()
		if application:
			application.add_window(icon)

	def teardown(self):
		global GLOBAL_TRAYICON
		self.window.hideonclose = False
		if GLOBAL_TRAYICON:
			GLOBAL_TRAYICON.destroy()
			GLOBAL_TRAYICON = None


class TrayIconBase(object):
	'''Base class for the zim tray icon.
	Contains code to create the tray icon menus.
	'''

	def get_application(self):
		return Gio.Application.get_default()

	def get_trayicon_menu(self):
		'''Returns the main 'tray icon menu'''
		menu = Gtk.Menu()

		item = Gtk.MenuItem.new_with_mnemonic(_('_Quick Note...')) # T: menu item in tray icon menu
		item.connect_object('activate', self.__class__.do_quick_note, self)
		menu.append(item)

		menu.append(Gtk.SeparatorMenuItem())

		notebooks = self.list_all_notebooks()
		self.populate_menu_with_notebooks(menu, notebooks)

		item = Gtk.MenuItem.new_with_mnemonic('  ' + _('_Other...'))  # Hack - using '  ' to indent visually
			# T: menu item in tray icon menu
		item.connect_object('activate', self.__class__.do_open_notebook, self)
		menu.append(item)

		menu.append(Gtk.SeparatorMenuItem())

		item = Gtk.MenuItem.new_with_mnemonic(_('_Quit')) # T: menu item in tray icon menu
		item.connect_object('activate', self.__class__.do_quit, self)
		menu.append(item)

		return menu

	def list_open_notebooks(self):
		'''Returns a list of open notebook.

		This method is to be implemented in child classes.

		@returns: a list of L{NotebookInfo} objects
		'''
		return [] # TODO
		#~ for uri in self.server.list_notebooks():
			#~ info = NotebookInfo(uri)
			#~ info.active = True
			#~ yield info

	def list_all_notebooks(self):
		'''Returns a list of all notebooks known in the current context

		This method mixes notebooks from L{list_open_notebooks()} with
		input from L{get_notebook_list()}. Open notebooks will have
		the C{active} attribute set.

		@returns: a list of L{NotebookInfo} objects
		'''
		uris = set()
		notebooks = [info for info in get_notebook_list()]
		for info in self.list_open_notebooks():
			if info in notebooks:
				# info from notebook list is updated already
				i = notebooks.index(info)
				notebooks[i].active = True
			else:
				info.update()
				info.active = True
				notebooks.append(info)

		for info in notebooks:
			if not info.active:
				info.active = False # None -> False

		return notebooks

	def populate_menu_with_notebooks(self, menu, notebooks):
		'''Populate a menu with a list of notebooks'''
		# TODO put checkbox behind open notebooks ?
		item = Gtk.MenuItem.new_with_mnemonic(_('Notebooks')) # T: menu item in tray icon menu
		item.set_sensitive(False)
		menu.append(item)

		if isinstance(notebooks, NotebookInfoList):
			notebooks = [info for info in notebooks] # copy

		notebooks.sort(key=lambda info: info.name)

		for info in notebooks:
			#~ print('>>>', info)
			item = Gtk.MenuItem.new_with_mnemonic('  ' + info.name)
				# Hack - using '  ' to indent visually
			if info.active:
				child = item.get_child()
				if isinstance(child, Gtk.Label):
					# FIXME this doesn't seem to work in Ubuntu menu :(
					child.set_markup('  <b>' + info.name + '</b>')
						# Hack - using '  ' to indent visually
			item.connect('activate', lambda o, u: self.do_activate_notebook(u), info.uri)
			menu.append(item)

	def do_activate_notebook(self, uri):
		'''Open a specific notebook.'''
		self.get_application().open_notebook(uri)

	def do_quit(self):
		'''Quit zim.'''
		self.get_application().quit()

	def do_open_notebook(self):
		'''Opens the notebook dialogs'''
		from zim.gui.notebookdialog import NotebookDialog
		NotebookDialog.unique(self, None, callback=self.do_activate_notebook).show()

	def do_quick_note(self):
		'''Show the dialog from the quicknote plugin'''
		from zim.plugins.quicknote import QuickNoteDialog
		dialog = QuickNoteDialog(None)
		dialog.show_all()


class StatusIconTrayIcon(TrayIconBase, Gtk.StatusIcon):
	'''Base class for a tray icon based on Gtk.StatusIcon'''

	__gsignals__ = {
		'destroy': (GObject.SignalFlags.RUN_LAST, None, ())
	}

	def __init__(self):
		Gtk.StatusIcon.__init__(self)

		icon_theme = Gtk.IconTheme.get_default()
		if icon_theme.has_icon('zim-panel'):
			self.set_from_icon_name('zim-panel')
		else:
			icon = data_file('zim.png').path
			self.set_from_file(icon)

		self.set_tooltip_text(_('Zim Desktop Wiki')) # T: tooltip for tray icon
		self.connect('popup-menu', self.__class__.do_popup_menu)

	def do_activate(self):
		open_notebooks = list(self.list_open_notebooks())
		if len(open_notebooks) == 0:
			# No open notebooks, open default or prompt full list
			notebooks = get_notebook_list()
			if notebooks.default:
				self.do_activate_notebook(notebooks.default.uri)
			else:
				self.do_popup_menu_notebooks(notebooks)
		elif len(open_notebooks) == 1:
			# Only one open notebook - present it
			self.do_activate_notebook(open_notebooks[0].uri)
		else:
			# Let the user choose from the open notebooks
			self.do_popup_menu_notebooks(open_notebooks)

	def do_popup_menu_notebooks(self, list, button=1, activate_time=0):
		menu = Gtk.Menu()
		self.populate_menu_with_notebooks(menu, list)
		menu.show_all()
		menu.popup(None, None, None, self, button, activate_time)

	def do_popup_menu(self, button=3, activate_time=0):
		#~ print('>>', button, activate_time)
		menu = self.get_trayicon_menu()
		menu.show_all()
		menu.popup(None, None, None, self, button, activate_time)

	def destroy(self):
		self.set_property('visible', False)
		self.emit('destroy')


_GLOBAL_INDICATOR = None
# This item is global, because if we create one, we can not detroy it
# again and it remains active in the desktop as an external resource


class AppIndicatorTrayIcon(TrayIconBase, SignalEmitter):
	'''Trayicon using the daemon and based on appindicator'''

	__signals__ = {
		'destroy': (None, None, ())
	}

	def __init__(self):
		# Note that even though we specify the icon "zim", the
		# ubuntu appindicator framework will first check for an icon
		# "zim-panel". This way it finds the mono color icons.
		global _GLOBAL_INDICATOR

		if not _GLOBAL_INDICATOR:
			_GLOBAL_INDICATOR = AppIndicator.Indicator.new(
				'zim-desktop-wiki', 'zim',
				AppIndicator.IndicatorCategory.APPLICATION_STATUS
			)

		self.appindicator = _GLOBAL_INDICATOR
		self.appindicator.set_status(AppIndicator.IndicatorStatus.ACTIVE)
		self.on_notebook_list_changed()

	def on_notebook_list_changed(self): # TODO connect somewhere
		menu = self.get_trayicon_menu()
		menu.show_all()
		self.appindicator.set_menu(menu)

	def destroy(self):
		self.appindicator.set_status(AppIndicator.IndicatorStatus.PASSIVE)
		self.emit('destroy')




# Copyright 2013-2020 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''Action interface classes.

Objects can have "actions", which are basically attributes of the
class L{Action} or L{ToggleAction}. These objects are callable as bound
methods. So actions are kind of special methods that define some
interface parameters, like what icon and label to use in the menu.

Use the L{action} and L{toggle_action} decorators to create actions.

The classes defined here can cooperate with C{Gio.Action} to tie into the
Gtk action framework. Also they can use Gtk widgets like C{Gtk.Button} as
a "proxy" to trigger the action and reflect the state.

## Menuhints

The "menuhints" attribute for actions sets one or more hints of where the action
should be in the menu and the behavior of the action. Multiple hints can
be separated with ":" in the string. The first one determines the menu, other
can modify the behavior.

Known values include:

  - notebook -- notebook section in "File" menu
  - page -- page section in "File" menu
  - edit -- "Edit" menu - modifies page, insensitive for read-only page
  - insert -- "Insert" menu & editor actionbar - modifies page, insensitive for read-only page
  - view -- "View" menu
  - tools -- "Tools" menu - also shown in toolbar plugin if an icon is provided and tool and not the "headerbar" hint
  - go -- "Go" menu
  - accelonly -- do not show in menu, shortcut key only
  - headerbar -- place action in the headerbar of the window, will place "view"
    menu items on the right, others on the left
  - toolbar -- used by toolbar plugin

Other values are ignored silently

 TODO: find right place in the documentation for this and update list

'''

import inspect
import weakref
import logging
import re

import zim.errors

from zim.signals import SignalHandler


logger = logging.getLogger('zim')

try:
	import gi
	gi.require_version('Gtk', '3.0')
	from gi.repository import Gtk
	from gi.repository import Gio
	from gi.repository import GLib
except:
	Gtk = None
	Gio = None
	GLib = None

def _get_modifier_mask():
	assert Gtk
	x, mod = Gtk.accelerator_parse('<Primary>')
	return mod

PRIMARY_MODIFIER_STRING = '<Primary>'
PRIMARY_MODIFIER_MASK = _get_modifier_mask()


def hasaction(obj, actionname):
	'''Like C{hasattr} but for attributes that define an action'''
	actionname = actionname.replace('-', '_')
	return hasattr(obj.__class__, actionname) \
		and isinstance(getattr(obj.__class__, actionname), ActionDescriptor)


class ActionDescriptor(object):

	_bound_class = None

	def __get__(self, instance, klass):
		if instance is None:
			return self # class access
		else:
			if instance not in self._bound_actions:
				self._bound_actions[instance] = self._bound_class(instance, self)
			return self._bound_actions[instance]


def action(label, accelerator='', icon=None, verb_icon=None, menuhints='', alt_accelerator=None, tooltip=None):
	'''Decorator to turn a method into an L{ActionMethod} object
	Methods decorated with this decorator can have keyword arguments
	but no positional arguments.
	@param label: the label used e.g for the menu item (can use "_" for mnemonics)
	@param accelerator: accelerator key description
	@param icon: name of a "noun" icon - used together with the label. Only use
	this for "things and places", not for actions or commands, and only if the
	icon makes the item easier to recognize.
	@param verb_icon: name of a "verb" icon - only used for compact menu views
	@param menuhints: string with hints for menu placement and sensitivity
	@param alt_accelerator: alternative accelerator key binding
	@param tooltip: tooltip label, defaults to C{label}
	'''
	def _action(function):
		return ActionClassMethod(function.__name__, function, label, icon, verb_icon, accelerator, alt_accelerator, menuhints, tooltip)

	return _action


class BoundActionMethod(object):

	def __init__(self, instance, action):
		self._instance = instance
		self._action = action
		self._sensitive = True
		self._proxies = set()
			# NOTE: Wanted to use WeakSet() here, but somehow we loose refs to
			#       widgets still being displayed
		self._gaction = None

	def __call__(self, *args, **kwargs):
		if not self._sensitive:
			raise AssertionError('Action not senitive: %s' % self.name)
		return self._action.func(self._instance, *args, **kwargs)

	def __getattr__(self, name):
		return getattr(self._action, name)

	def get_sensitive(self):
		return self._sensitive

	def set_sensitive(self, sensitive):
		self._sensitive = sensitive

		if self._gaction:
			self._gaction.set_enabled(sensitive)

		for proxy in self._proxies:
			proxy.set_sensitive(sensitive)

	def get_gaction(self):
		if self._gaction is None:
			assert Gio is not None
			self._gaction = Gio.SimpleAction.new(self.name)
			self._gaction.set_enabled(self._sensitive)
			self._gaction.connect('activate', self._on_activate)
		return self._gaction

	def _on_activate(self, proxy, value):
		# "proxy" can either be Gtk.Button, Gtk.Action or Gio.Action
		logger.debug('Action: %s', self.name)
		try:
			self.__call__()
		except:
			zim.errors.exception_handler(
				'Exception during action: %s' % self.name)

	def _connect_gtkaction(self, gtkaction):
		gtkaction.connect('activate', self._on_activate_proxy)
		gtkaction.set_sensitive(self._sensitive)
		self._proxies.add(gtkaction)


class ActionMethod(BoundActionMethod):

	_button_class = Gtk.Button if Gtk is not None else None
	_tool_button_class = Gtk.ToolButton if Gtk is not None else None

	def create_button(self):
		assert Gtk is not None
		button = self._button_class.new_with_mnemonic(self.label)
		button.set_tooltip_text(self.tooltip)
		self.connect_button(button)
		return button

	def create_icon_button(self, fallback_icon=None):
		assert Gtk is not None
		icon_name = self.verb_icon or self.icon or fallback_icon
		assert icon_name, 'No icon or verb_icon defined for action "%s"' % self.name
		icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
		button = self._button_class()
		button.set_image(icon)
		button.set_tooltip_text(self.tooltip) # icon button should always have tooltip
		self.connect_button(button)
		return button

	def create_tool_button(self, fallback_icon=None, connect_button=True):
		assert Gtk is not None
		icon_name = self.verb_icon or self.icon or fallback_icon
		assert icon_name, 'No icon or verb_icon defined for action "%s"' % self.name
		button = self._tool_button_class()
		button.set_label(self.label)
		button.set_use_underline(True)
		button.set_icon_name(icon_name)
		button.set_tooltip_text(self.tooltip) # icon button should always have tooltip

		if connect_button:
			self.connect_button(button)
		return button

	def connect_button(self, button):
		button.connect('clicked', self._on_activate_proxy)
		button.set_sensitive(self._sensitive)
		self._proxies.add(button)
		button.connect('destroy', self._on_destroy_proxy)

	def _on_destroy_proxy(self, proxy):
		self._proxies.discard(proxy)

	def _on_activate_proxy(self, proxy):
		self._on_activate(proxy, None)


class ActionClassMethod(ActionDescriptor):

	_bound_class = ActionMethod
	_n_args = 1 # self

	def __init__(self, name, func, label, icon=None, verb_icon=None, accelerator='', alt_accelerator=None, menuhints='', tooltip=None):
		assert self._assert_args(func), '%s() has incompatible argspec' % func.__name__
		tooltip = tooltip or label.replace('_', '')
		self.name = name
		self.func = func
		self.label = label
		self.tooltip = tooltip
		self.icon = icon
		self.verb_icon = verb_icon
		self.hasicon = bool(self.verb_icon or self.icon)
		self.menuhints = menuhints.split(':')

		self._attr = (self.name, label, tooltip, icon or verb_icon)
		self._alt_attr = (self.name + '_alt1', label, tooltip, icon or verb_icon)
		self._accel = accelerator
		self._alt_accel = alt_accelerator

		self._bound_actions = weakref.WeakKeyDictionary()

	def _assert_args(self, func):
		spec = inspect.getfullargspec(func)
		if spec.defaults:
			return len(spec.defaults) == len(spec.args) - self._n_args
		else:
			return len(spec.args) == self._n_args


def toggle_action(label, accelerator='', icon=None, verb_icon=None, init=False, menuhints='', tooltip=None):
	'''Decorator to turn a method into an L{ToggleActionMethod} object

	The decorated method should be defined as:
	C{my_toggle_method(self, active)}. The 'C{active}' parameter is a
	boolean that reflects the new state of the toggle.

	Users can also call the method without setting the C{active}
	parameter. In this case the wrapper determines how to toggle the
	state and calls the inner function with the new state.

	@param label: the label used e.g for the menu item (can use "_" for mnemonics)
	@param accelerator: accelerator key description
	@param icon: name of a "noun" icon - used together with the label. Only use
	this for "things and places", not for actions or commands, and only if the
	icon makes the item easier to recognize.
	@param verb_icon: name of a "verb" icon - only used for compact menu views
	@param init: initial state of the toggle
	@param menuhints: string with hints for menu placement and sensitivity
	@param tooltip: tooltip label, defaults to C{label}
	'''
	def _toggle_action(function):
		return ToggleActionClassMethod(function.__name__, function, label, icon, verb_icon, accelerator, init, menuhints, tooltip)

	return _toggle_action


class ToggleActionMethod(ActionMethod):

	_button_class = Gtk.ToggleButton if Gtk is not None else None
	_tool_button_class = Gtk.ToggleToolButton if Gtk is not None else None

	def __init__(self, instance, action):
		ActionMethod.__init__(self, instance, action)
		self._state = action._init

	def __call__(self, active=None):
		if not self._sensitive:
			raise AssertionError('Action not senitive: %s' % self.name)

		if active is None:
			active = not self._state
		elif active == self._state:
			return # nothing to do

		with self._on_activate.blocked():
			self._action.func(self._instance, active)
		self.set_active(active)

	def create_tool_button(self, fallback_icon=None, connect_button=True):
		if connect_button:
			raise NotImplementedError # Should work but gives buggy behavior, try using gaction + set_action_name() instead
		return ActionMethod.create_tool_button(self, fallback_icon, connect_button)

	def connect_button(self, button):
		'''Connect a C{Gtk.ToggleAction} or C{Gtk.ToggleButton} to this action'''
		button.set_active(self._state)
		button.set_sensitive(self._sensitive)
		button.connect('toggled', self._on_activate_proxy)
		self._proxies.add(button)

	_connect_gtkaction = connect_button

	def _on_activate_proxy(self, proxy):
		self._on_activate(proxy, proxy.get_active())

	@SignalHandler
	def _on_activate(self, proxy, active):
		'''Callback for activate signal of connected objects'''
		if active != self._state:
			logger.debug('Action: %s(%s)', self.name, active)
			try:
				self.__call__(active)
			except Exception as error:
				zim.errors.exception_handler(
					'Exception during toggle action: %s(%s)' % (self.name, active))

	def get_active(self):
		return self._state

	def set_active(self, active):
		'''Change the state of the action without triggering the action'''
		if active == self._state:
			return
		self._state = active

		if self._gaction:
			self._gaction.set_state(GLib.Variant.new_boolean(self._state))

		with self._on_activate.blocked():
			for proxy in self._proxies:
				if isinstance(proxy, Gtk.ToggleToolButton):
					pass
				else:
					proxy.set_active(active)

	def get_gaction(self):
		if self._gaction is None:
			assert Gio is not None
			self._gaction = Gio.SimpleAction.new_stateful(self.name, None, GLib.Variant.new_boolean(self._state))
			self._gaction.set_enabled(self._sensitive)
			self._gaction.connect('activate', self._on_activate)
		return self._gaction


class ToggleActionClassMethod(ActionClassMethod):
	'''Toggle action, used by the L{toggle_action} decorator'''

	_bound_class = ToggleActionMethod
	_n_args = 2 # self, active

	def __init__(self, name, func, label, icon=None, verb_icon=None, accelerator='', init=False, menuhints='', tooltip=None):
		# The ToggleAction instance lives in the client class object;
		# using weakkeydict to store instance attributes per
		# client object
		ActionClassMethod.__init__(self, name, func, label, icon, verb_icon, accelerator, menuhints=menuhints, tooltip=tooltip)
		self._init = init

	def _assert_args(self, func):
		spec = inspect.getfullargspec(func)
		return len(spec.args) == 2 # (self, active)


def radio_action(menulabel, *radio_options, menuhints=''):
	def _action(function):
		return RadioActionClassMethod(function.__name__, function, menulabel, radio_options, menuhints)

	return _action


def radio_option(key, label, accelerator=''):
	tooltip = label.replace('_', '')
	return (key, None, label, accelerator, tooltip)
		# tuple must match spec for actiongroup.add_radio_actions()


def gtk_radioaction_set_current(g_radio_action, key):
	# Gtk.radioaction.set_current is gtk >= 2.10
	for a in g_radio_action.get_group():
		if a.get_name().endswith('_' + key):
			a.activate()
			break


class RadioActionMethod(BoundActionMethod):

	def __init__(self, instance, action):
		BoundActionMethod.__init__(self, instance, action)
		self._state = None

	def __call__(self, key):
		if not key in self.keys:
			raise ValueError('Invalid key: %s' % key)
		self.func(self._instance, key)
		self.set_state(key)

	def get_state(self):
		return self._state

	def set_state(self, key):
		self._state = key
		for proxy in self._proxies:
			gtk_radioaction_set_current(proxy, key)

	def get_gaction(self):
		raise NotImplementedError # TODO

	def _connect_gtkaction(self, gtkaction):
		gtkaction.connect('changed', self._on_gtkaction_changed)
		self._proxies.add(gtkaction)
		if self._state is not None:
			gtk_radioaction_set_current(gtkaction, self._state)

	def _on_gtkaction_changed(self, gaction, current):
		try:
			name = current.get_name()
			assert name.startswith(self.name + '_')
			key = name[len(self.name) + 1:]
			if self._state == key:
				pass
			else:
				logger.debug('Action: %s(%s)', self.name, key)
				self.__call__(key)
		except:
			zim.errors.exception_handler(
				'Exception during action: %s(%s)' % (self.name, key))


class RadioActionClassMethod(ActionDescriptor):

	_bound_class = RadioActionMethod

	def __init__(self, name, func, menulabel, radio_options, menuhints=''):
		self.name = name
		self.func = func
		self.menulabel = menulabel
		self.keys = [opt[0] for opt in radio_options]
		self._entries = tuple(
			(name + '_' + opt[0],) + opt[1:] + (i,)
				for i, opt in enumerate(radio_options)
		)
		self.hasicon = False
		self.menuhints = menuhints.split(':')
		self._bound_actions = weakref.WeakKeyDictionary()


def get_actions(obj):
	'''Returns bound actions for object

	NOTE: See also L{zim.plugins.list_actions()} if you want to include actions
	of plugin extensions
	'''
	actions = []
	for name, action in inspect.getmembers(obj.__class__, lambda m: isinstance(m, ActionDescriptor)):
		actions.append((name, action.__get__(obj, obj.__class__)))
	return actions


def get_gtk_actiongroup(obj):
	'''Return a C{Gtk.ActionGroup} for an object using L{Action}
	objects as attributes.

	Defines the attribute C{obj.actiongroup} if it does not yet exist.

	This method can only be used when gtk is available
	'''
	assert Gtk is not None

	if hasattr(obj, 'actiongroup') \
	and obj.actiongroup is not None:
		return obj.actiongroup

	obj.actiongroup = Gtk.ActionGroup(obj.__class__.__name__)

	for name, action in get_actions(obj):
		if isinstance(action, RadioActionMethod):
			obj.actiongroup.add_radio_actions(action._entries)
			gtkaction = obj.actiongroup.get_action(action._entries[0][0])
			action._connect_gtkaction(gtkaction)
		else:
			_gtk_add_action_with_accel(obj, obj.actiongroup, action, action._attr, action._accel)
			if action._alt_accel:
				_gtk_add_action_with_accel(obj, obj.actiongroup, action, action._alt_attr, action._alt_accel)

	return obj.actiongroup


def _gtk_add_action_with_accel(obj, actiongroup, action, attr, accel):
	assert Gtk is not None

	if isinstance(action, ToggleActionMethod):
		gtkaction = Gtk.ToggleAction(*attr)
	else:
		gtkaction = Gtk.Action(*attr)

	gtkaction.zim_readonly = not bool(
		'edit' in action.menuhints or 'insert' in action.menuhints
	)
	action._connect_gtkaction(gtkaction)
	actiongroup.add_action_with_accel(gtkaction, accel)


def initialize_actiongroup(obj, prefix):
	assert Gio is not None
	actiongroup = Gio.SimpleActionGroup()
	for name, action in get_actions(obj):
		gaction = action.get_gaction()
		actiongroup.add_action(gaction)
	obj.insert_action_group(prefix, actiongroup)



#
# Copyright 2010 Thorsten Hackbarth <thorsten.hackbarth@gmx.de>
#           2011-2015 Jaap Karssenberg <jaap.karssenberg@gmail.com>
# License:  same as zim (gpl)
#
#
# !! NOTE: when changing this plugin, do test performance on a folder with lots of photos!
#
# ChangeLog
# 2021-01-16 Added rename and delete functionality to file browsers context menu (Thomas Engel)
# 2015-10-04 Reworked threads to better avoid blocking user interface (Jaap)
# 2013-03-03 Change to new plugin extension structure (Jaap)
# 2013-02-25 Added zooming icon size, made icon rendering more robust (Jaap)
# 2012-08-17 Added mimetype icons, statusbar toggle button, and gio monitor support (Jaap)
# 2012-07-20 Updated code for pane uistate (Jaap)
# 2012-04-17 Allow drag&drop when folder does not exist yet + fix drag&drop on windows (Jaap)
# 2012-02-29 Further work on making iconview look nice and support drag&drop (Jaap)
# 2012-02-27 Complete refactoring of thumbnail manager + test case (Jaap)
# 2012-02-26 Rewrote direct filessystem calls in order to support non-utf8 file systems (Jaap)
# 2011-01-25 Refactored widget and plugin code (Jaap)
#		tested on gtk < 2.12 (tooltip interface)
#		add pref for image magick (convert cmd exists on win32 but is not the same)
#		added buttons to side of widget
# 2011-01-02 Fixed use of uistate and updated for new framework to add to the mainwindow (Jaap)
# 2010-11-14 Fixed Bug 664551
# 2010-08-31 freedesktop.org thumbnail spec mostly implemented
# 2010-06-29 1st working version
#
# TODO:
# [ ] Copy / cut files in context menu
# [ ] Button to clean up the folder - only show when the folder is empty
# [ ] Avoid scaling up small images when thumbnailing (check spec on this)
# [ ] Store failures as well


'''Zim plugin to display files in attachments folder.'''

import logging

from gi.repository import Gtk
from gi.repository import GObject


from zim.plugins import PluginClass
from zim.actions import toggle_action

from zim.gui.pageview import PageViewExtension
from zim.gui.applications import open_folder_prompt_create

from zim.gui.widgets import RIGHT_PANE, PANE_POSITIONS, \
	IconButton, ScrolledWindow, StatusPage, \
	WindowSidePaneWidget, uistate_property


from .filebrowser import FileBrowserIconView, THUMB_SIZE_LARGE


logger = logging.getLogger('zim.plugins.attachmentbrowser')


MIN_ICON_ZOOM = 16
DEFAULT_ICON_ZOOM = 64



class AttachmentBrowserPlugin(PluginClass):

	plugin_info = {
		'name': _('Attachment Browser'), # T: plugin name
		'description': _('''\
This plugin shows the attachments folder of the current page as an
icon view at bottom pane.
'''), # T: plugin description
		'author': 'Thorsten Hackbarth <thorsten.hackbarth@gmx.de>\n'
			      'Jaap Karssenberg <jaap.karssenberg@gmail.com>\n'
				  'Thomas Engel <realdatenwurm@gmail.com>',
		'help': 'Plugins:Attachment Browser',
	}

	plugin_preferences = (
		# key, type, label, default
		('pane', 'choice', _('Position in the window'), RIGHT_PANE, PANE_POSITIONS),
			# T: option for plugin preferences
		('use_thumbnails', 'bool', _('Use thumbnails'), True),
			# T: option for plugin preferences
		('thumbnail_svg', 'bool', _('Support thumbnails for SVG'), False),
			# T: option for plugin preferences
			# NOTE: svg cases crashes on some systems, so needs to be off by default

	#	('preview_size', 'int', _('Tooltip preview size [px]'), (THUMB_SIZE_MIN,480,THUMB_SIZE_MAX)), # T: input label
	#	('thumb_quality', 'int', _('Preview jpeg Quality [0..100]'), (0,50,100)), # T: input label
	#~	('use_imagemagick', 'bool', _('Use ImageMagick for thumbnailing'), False), # T: input label
	)

	#~ @classmethod
	#~ def check_dependencies(klass):
		#~ return [("ImageMagick",Application(('convert',None)).tryexec())]


class AttachmentBrowserWindowExtension(PageViewExtension):

	def __init__(self, plugin, window):
		PageViewExtension.__init__(self, plugin, window)
		self.preferences = plugin.preferences
		self._monitor = None

		# Init browser widget
		self.widget = AttachmentBrowserPluginWidget(self, self.navigation, self.preferences)

		if self.pageview.page is not None:
			self.on_page_changed(self.pageview, self.pageview.page)
		self.connectto(self.pageview, 'page-changed')

		self.add_sidepane_widget(self.widget, 'pane')

	def on_page_changed(self, pageview, page):
		self.widget.set_folder(
			pageview.notebook.get_attachments_dir(page)
		)

	def teardown(self):
		self.widget.iconview.teardown_folder()


class AttachmentBrowserPluginWidget(Gtk.Box, WindowSidePaneWidget):
	'''Wrapper aroung the L{FileBrowserIconView} that adds the buttons
	for zoom / open folder / etc. ...
	'''

	title = _('Atta_chments') # T: label for attachment browser pane

	icon_size = uistate_property('icon_size', DEFAULT_ICON_ZOOM)

	def __init__(self, extension, opener, preferences):
		Gtk.Box.__init__(self, Gtk.Orientation.VERTICAL)
		self.set_homogeneous(False)
		self.extension = extension # XXX
		self.opener = opener
		self.uistate = extension.uistate
		self.preferences = preferences
		self._close_button = None

		self.iconview = FileBrowserIconView(opener, self.icon_size)
		self._stack = Gtk.Stack()
		for name, widget in (
			('iconview', ScrolledWindow(self.iconview, shadow=Gtk.ShadowType.NONE)),
			('placeholder', StatusPage('folder-symbolic', _('No attachments'))), # T: placeholder label for sidepane
		):
			widget.show_all()
			self._stack.add_named(widget, name)
		self._stack.set_visible_child_name('placeholder')

		self.pack_start(self._stack, True, True, 0)

		self.on_preferences_changed()
		self.preferences.connect('changed', self.on_preferences_changed)

		self.buttonbox = Gtk.Box(Gtk.Orientation.HORIZONTAL)
		self.pack_end(self.buttonbox, False, True, 0)

		open_folder_button = IconButton(Gtk.STOCK_OPEN, relief=False)
		open_folder_button.connect('clicked', self.on_open_folder)
		self.buttonbox.pack_start(open_folder_button, False, True, 0)

		refresh_button = IconButton(Gtk.STOCK_REFRESH, relief=False)
		refresh_button.connect('clicked', lambda o: self.on_refresh_button())
		self.buttonbox.pack_start(refresh_button, False, True, 0)

		zoomin = IconButton(Gtk.STOCK_ZOOM_IN, relief=False)
		zoomout = IconButton(Gtk.STOCK_ZOOM_OUT, relief=False)
		zoomin.connect('clicked', lambda o: self.on_zoom_in())
		zoomout.connect('clicked', lambda o: self.on_zoom_out())
		self.buttonbox.pack_end(zoomout, False, True, 0)
		self.buttonbox.pack_end(zoomin, False, True, 0)
		self.zoomin_button = zoomin
		self.zoomout_button = zoomout

		self.set_icon_size(self.icon_size)

		self.iconview.connect('folder-changed', lambda o: self.update_status())

	def set_orientation(self, orientation):
		Gtk.Box.set_orientation(self, orientation)
		self.buttonbox.set_orientation(
			Gtk.Orientation.VERTICAL if orientation is Gtk.Orientation.HORIZONTAL else Gtk.Orientation.HORIZONTAL
		) # reverse of widget orientation

	def on_preferences_changed(self, *a):
		self.iconview.set_use_thumbnails(self.preferences['use_thumbnails'])
		self.iconview.set_thumbnail_svg(self.preferences['thumbnail_svg'])

	def set_folder(self, folder):
		self.iconview.set_folder(folder)
		self.update_status()

	def update_status(self):
		n = len(self.iconview.get_model())
		self.set_info(ngettext('%i Atta_chment', '%i Atta_chments', n) % n)
		# T: Label for the statusbar, %i is the number of attachments for the current page

		if n == 0:
			self._stack.set_visible_child_name('placeholder')
			self.zoomin_button.set_sensitive(False)
			self.zoomout_button.set_sensitive(False)
		else:
			self._stack.set_visible_child_name('iconview')
			self.zoomin_button.set_sensitive(self.icon_size < THUMB_SIZE_LARGE)
			self.zoomout_button.set_sensitive(self.icon_size > MIN_ICON_ZOOM)

	def set_embeded_closebutton(self, button):
		if self._close_button:
			self.buttonbox.remove(self._close_button)

		if self.get_orientation() == Gtk.Orientation.VERTICAL:
			return False

		if button is not None:
			self.buttonbox.pack_start(button, False, True, 0)
			self.buttonbox.reorder_child(button, 0)

		self._close_button = button
		return True

	def on_open_folder(self, o):
		# Callback for the "open folder" button
		open_folder_prompt_create(self, self.iconview.folder)
		self.iconview.refresh()

	def on_refresh_button(self):
		self.iconview.refresh()
		self.update_status()

	def on_zoom_in(self):
		self.set_icon_size(
			min((self.icon_size * 2, THUMB_SIZE_LARGE))) # 16 > 32 > 64 > 128 > 256

	def on_zoom_out(self):
		self.set_icon_size(
			max((self.icon_size / 2, MIN_ICON_ZOOM))) # 16 < 32 < 64 < 128 < 256

	def set_icon_size(self, icon_size):
		self.iconview.set_icon_size(icon_size)
		self.zoomin_button.set_sensitive(icon_size < THUMB_SIZE_LARGE)
		self.zoomout_button.set_sensitive(icon_size > MIN_ICON_ZOOM)
		self.icon_size = icon_size # Do this last - avoid store state after fail
